from methoddispatch import singledispatch

from melange.domain_event_bus import DomainEventHandler


class Debugger(DomainEventHandler):
    @singledispatch
    def process(self, event):
        print(f"Event received: {type(event).__name__} at {event.occurred_on.isoformat()}")



from methoddispatch import SingleDispatch, singledispatch

from melange.messaging import ExchangeListener


class CommandHandler(SingleDispatch):
    @singledispatch
    def execute(self, command):
        pass


command_handler = CommandHandler.execute.register


class QueryHandler(SingleDispatch):
    @singledispatch
    def execute(self, command):
        pass


query_handler = QueryHandler.execute.register


class Command:
    pass


class Query:
    pass


class AsyncReadModel(ExchangeListener):
    pass


class WorkflowHandler(ExchangeListener):
    pass



import json

from melange.drivers.aws import parse_event_from_sns


class TestUtils:

    def test_parse_event_from_sns(self):
        sns_event = {
            'Records': [
                {
                    'Sns': {
                        'Message': json.dumps({'event_type_name': 'some_name', 'value': 'some_message'})
                    }
                }
            ]
        }

        parsed_event = parse_event_from_sns(sns_event)

        assert parsed_event['value'] == 'some_message'



from datetime import datetime

import pytz


class DomainEvent:
    def __init__(self, occurred_on=None):
        self.occurred_on = occurred_on or datetime.now(pytz.utc)

    def get_occurred_on(self):
        return self.occurred_on

    def emit(self, **kwargs):
        from melange.domain_event_bus import DomainEventBus
        DomainEventBus().publish(self, **kwargs)
        return self



import logging
import threading

from singleton import Singleton

from .domain_event import DomainEvent
from .domain_subscriber import DomainEventHandler


logger = logging.getLogger(__name__)


class DomainEventBus(metaclass=Singleton):
    thread_local = threading.local()

    def __init__(self):
        self.thread_local.subscribers = []
        self.thread_local.publishing = False
        self.thread_local.keep_publishing = False

    def is_publishing(self):
        return getattr(self.thread_local, 'publishing', False)

    def keep_publishing(self):
        self.thread_local.keep_publishing = True

    def should_keep_publishing(self):
        return getattr(self.thread_local, 'keep_publishing', False)

    def publish(self, event, **kwargs):
        if self.is_publishing() and not self.should_keep_publishing():
            return

        if not isinstance(event, DomainEvent):
            logger.error('Invalid data passed. You must pass an event instance')
            return

        self.thread_local.publishing = True
        self._publish_synchronous(event, **kwargs)
        self.thread_local.publishing = False

    def reset(self):
        val = getattr(self.thread_local, 'publishing', None)
        if val is None:
            val = False
            self.thread_local.publishing = False
            self.thread_local.keep_publishing = False

        if not self.is_publishing():
            self.thread_local.subscribers = []

    def subscribe(self, subscriber):
        if self.is_publishing() or not isinstance(subscriber, DomainEventHandler):
            return False

        if subscriber not in self.thread_local.subscribers:
            self.thread_local.subscribers.append(subscriber)

    def _publish_synchronous(self, event, **kwargs):
        subscribers = self._get_subscribers(event)

        for subscr in subscribers:
            try:
                event._meta = kwargs
                subscr.process(event)
            except Exception as e:
                logger.exception(e)

    def _get_subscribers(self, event):
        return [subscriber for subscriber in self.thread_local.subscribers if subscriber.accepts(event)]

    def get_publishing(self):
        val = getattr(self.thread_local, 'publishing', None)
        if val is None:
            self.thread_local.publishing = False

        return self.thread_local.publishing



""" Subscriber super-class """

from funcy import lfilter
from methoddispatch import SingleDispatch, singledispatch

from .domain_event_bus import DomainEvent


class DomainEventHandler(SingleDispatch):
    def process(self, event, **kwargs):
        """
        Called by the domain_event_bus.

        :param event: The event object
        :type event: EventMessage or subclass of event

        This method implements the logic for processing the event. This method should not block for
        long time as that will affect the performance of the domain_event_bus.
        """
        self._process(event, **kwargs)

    @singledispatch
    def _process(self, event, **kwargs):
        pass

    def listens_to(self):
        return lfilter(lambda t: t is not object, self._process.registry)

    def listen(self):
        from melange.domain_event_bus import DomainEventBus
        DomainEventBus().subscribe(self)
        return self

    def accepts(self, event):
        return (
            isinstance(event, DomainEvent)
            and (self.listens_to() == [] or any(isinstance(event, e) for e in self.listens_to()))
        )


event_handler = DomainEventHandler._process.register



import json
import uuid
from json import JSONDecodeError

import boto3

from melange.messaging import MessagingDriver, Message


class AWSDriver(MessagingDriver):
    def __init__(self, **kwargs):
        super().__init__()
        self.max_number_of_messages = kwargs.get('max_number_of_messages', 10)
        self.visibility_timeout = kwargs.get('visibility_timeout', 100)
        self.wait_time_seconds = kwargs.get('wait_time_seconds', 10)

    def declare_topic(self, topic_name):
        sns = boto3.resource('sns')
        topic = sns.create_topic(Name=topic_name)
        return topic

    def get_queue(self, queue_name):
        sqs_res = boto3.resource('sqs')

        return sqs_res.get_queue_by_name(QueueName=queue_name)

    def declare_queue(self, queue_name, *topics_to_bind, dead_letter_queue_name=None, **kwargs):
        try:
            queue = self.get_queue(queue_name)
        except Exception:
            queue = self._create_queue(queue_name, content_based_deduplication='true')

        if topics_to_bind:
            statements = []
            for topic in topics_to_bind:
                statement = {
                    'Sid': 'Sid{}'.format(uuid.uuid4()),
                    'Effect': 'Allow',
                    'Principal': '*',
                    'Resource': queue.attributes['QueueArn'],
                    'Action': 'sqs:SendMessage',
                    'Condition': {
                        'ArnEquals': {
                            'aws:SourceArn': topic.arn
                        }
                    }
                }

                statements.append(statement)
                subscription = topic.subscribe(Protocol='sqs', Endpoint=queue.attributes['QueueArn'])

                if kwargs.get('filter_events'):
                    filter_policy = {
                        'event_type': kwargs['filter_events']
                    }
                else:
                    filter_policy = {}

                subscription.set_attributes(AttributeName='FilterPolicy',
                                            AttributeValue=json.dumps(filter_policy))

            policy = {
                'Version': '2012-10-17',
                'Id': 'sqspolicy',
                'Statement': statements
            }

            queue.set_attributes(Attributes={'Policy': json.dumps(policy)})

        dead_letter_queue = None
        if dead_letter_queue_name:
            try:
                dead_letter_queue = self.get_queue(dead_letter_queue_name)
            except Exception:
                dead_letter_queue = self._create_queue(
                    dead_letter_queue_name, content_based_deduplication='true')

            redrive_policy = {
                'deadLetterTargetArn': dead_letter_queue.attributes['QueueArn'],
                'maxReceiveCount': '4'
            }

            queue.set_attributes(Attributes={'RedrivePolicy': json.dumps(redrive_policy)})

        return queue, dead_letter_queue

    def _create_queue(self, queue_name, **kwargs):
        sqs_res = boto3.resource('sqs')
        fifo = queue_name.endswith('.fifo')
        attributes = {}
        if fifo:
            attributes['FifoQueue'] = 'true'
            attributes['ContentBasedDeduplication'] = 'true' if kwargs.get('content_based_deduplication') else 'false'
        queue = sqs_res.create_queue(QueueName=queue_name, Attributes=attributes)
        return queue

    def retrieve_messages(self, queue, attempt_id=None):
        kwargs = dict(
            MaxNumberOfMessages=self.max_number_of_messages,
            VisibilityTimeout=self.visibility_timeout,
            WaitTimeSeconds=self.wait_time_seconds,
            AttributeNames=['All']
        )

        if attempt_id:
            kwargs['ReceiveRequestAttemptId'] = attempt_id

        messages = queue.receive_messages(**kwargs)

        return [Message(message.message_id, self._extract_message_content(message), message)
                for message in messages]

    def queue_publish(
            self, content, queue, event_type_name=None,
            message_group_id=None, message_deduplication_id=None):
        kwargs = dict(
            MessageBody=content
        )

        if event_type_name:
            kwargs['MessageAttributes'] = {
                'event_type': {
                    'DataType': 'String',
                    'StringValue': event_type_name
                }
            }

        if message_group_id:
            kwargs['MessageGroupId'] = message_group_id

        if message_deduplication_id:
            kwargs['MessageDeduplicationId'] = message_deduplication_id

        queue.send_message(**kwargs)

    def publish(self, content, topic, event_type_name):
        response = topic.publish(Message=content, MessageAttributes={
            'event_type': {
                'DataType': 'String',
                'StringValue': event_type_name
            }
        })

        if 'MessageId' not in response:
            raise ConnectionError('Could not send the event to the SNS TOPIC')

    def acknowledge(self, message):
        message.metadata.delete()

    def close_connection(self):
        pass

    def delete_queue(self, queue):
        queue.delete()

    def delete_topic(self, topic):
        topic.delete()

    def _extract_message_content(self, message):
        body = message.body
        try:
            message_content = json.loads(body)
            if 'Message' in message_content:
                content = json.loads(message_content['Message'])
            else:
                content = message_content
        except JSONDecodeError:
            content = body

        return content



from threading import Thread

from melange.domain_event_bus import DomainEvent, DomainEventBus, DomainEventHandler
from melange.domain_event_bus.debugger import Debugger


class TestDomainEventBus:
    def test_publishing_and_receive_an_event(self):
        self.listened_event = None

        class TestDomainEvent(DomainEvent):
            pass

        class TestDomainSubscriber(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_event = event

            def listens_to(x):
                return [DomainEvent]

        test_domain_event = TestDomainEvent()
        DomainEventBus().reset()
        TestDomainSubscriber().listen()
        test_domain_event.emit()

        assert self.listened_event == test_domain_event

    def test_receive_all_event_types_if_listens_to_is_not_overriden(self):
        self.listened_events = []

        class TestDomainEvent(DomainEvent):
            pass

        class TestDomainEvent2(DomainEvent):
            pass

        class TestDomainSubscriber(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events.append(event)

        DomainEventBus().reset()
        TestDomainSubscriber().listen()
        test_domain_event_1 = TestDomainEvent().emit()
        test_domain_event_2 = TestDomainEvent2().emit()
        test_domain_event_3 = TestDomainEvent().emit()
        test_domain_event_4 = TestDomainEvent2().emit()

        assert self.listened_events[0] == test_domain_event_1
        assert self.listened_events[1] == test_domain_event_2
        assert self.listened_events[2] == test_domain_event_3
        assert self.listened_events[3] == test_domain_event_4

    def test_receive_only_specifically_listened_events(self):
        self.listened_events_1 = []
        self.listened_events_2 = []

        class TestDomainEvent(DomainEvent):
            name = 'TestDomainEvent'

        class TestDomainEvent2(DomainEvent):
            name = 'TestDomainEvent2'

        class TestDomainEvent3(DomainEvent):
            name = 'TestDomainEvent3'

        class TestDomainSubscriber(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_1.append(event)

            def listens_to(self):
                return [TestDomainEvent]

        class TestDomainSubscriber2(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_2.append(event)

            def listens_to(self):
                return [TestDomainEvent2]

        DomainEventBus().reset()
        TestDomainSubscriber().listen()
        TestDomainSubscriber2().listen()

        test_domain_event_1 = TestDomainEvent().emit()
        test_domain_event_2 = TestDomainEvent2().emit()
        test_domain_event_3 = TestDomainEvent().emit()
        TestDomainEvent3().emit()

        assert len(self.listened_events_1) == 2
        assert len(self.listened_events_2) == 1
        assert self.listened_events_1[0] == test_domain_event_1
        assert self.listened_events_1[1] == test_domain_event_3
        assert self.listened_events_2[0] == test_domain_event_2

    def test_the_domain_event_publisher_is_thread_bound(self):
        self.listened_events_1 = []
        self.listened_events_2 = []

        class TestDomainEvent(DomainEvent):
            name = 'TestDomainEvent'

        class TestDomainSubscriber(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_1.append(event)

            def listens_to(self):
                return [TestDomainEvent]

        class TestDomainSubscriber2(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_2.append(event)

            def listens_to(self):
                return [TestDomainEvent]

        def thread_1_func():
            DomainEventBus().reset()
            TestDomainSubscriber().listen()
            TestDomainEvent().emit()
            TestDomainEvent().emit()

        def thread_2_func():
            DomainEventBus().reset()
            TestDomainSubscriber2().listen()
            TestDomainEvent().emit()
            TestDomainEvent().emit()
            TestDomainEvent().emit()

        t1 = Thread(target=thread_1_func)
        t2 = Thread(target=thread_2_func)

        t1.start()
        t2.start()
        t1.join()
        t2.join()

        assert len(self.listened_events_1) == 2
        assert len(self.listened_events_2) == 3

    def test_the_bus_will_not_allow_chained_event_publishing(self):
        self.listened_events_1 = []
        self.listened_events_2 = []

        class TestDomainEvent(DomainEvent):
            name = 'TestDomainEvent'

        class TestDomainEvent2(DomainEvent):
            name = 'TestDomainEvent2'

        class TestDomainSubscriber(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_1.append(event)
                TestDomainEvent2().emit()

            def listens_to(self):
                return [TestDomainEvent]

        class TestDomainSubscriber2(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_2.append(event)

            def listens_to(self):
                return [TestDomainEvent2]

        DomainEventBus().reset()
        TestDomainSubscriber().listen()
        TestDomainSubscriber2().listen()
        TestDomainEvent().emit()

        assert len(self.listened_events_1) == 1
        assert len(self.listened_events_2) == 0

    def test_the_debugger_will_log_any_event(self):
        self.listened_events_1 = []
        self.listened_events_2 = []

        class TestDomainEvent(DomainEvent):
            name = 'TestDomainEvent'

        class TestDomainEvent2(DomainEvent):
            name = 'TestDomainEvent2'

        class TestDomainSubscriber(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_1.append(event)
                TestDomainEvent2().emit()

            def listens_to(self):
                return [TestDomainEvent]

        class TestDomainSubscriber2(DomainEventHandler):
            def process(x, event):
                nonlocal self
                self.listened_events_2.append(event)

            def listens_to(self):
                return [TestDomainEvent2]

        DomainEventBus().reset()
        TestDomainSubscriber().listen()
        TestDomainSubscriber2().listen()

        Debugger().listen()

        TestDomainEvent().emit()



from melange.drivers.aws import AWSDriver
from melange.drivers.rabbitmq_driver import RabbitMQDriver
from melange.messaging.driver_manager import DriverManager

DriverManager().add_available_drivers(
    aws=AWSDriver, rabbitMQ=RabbitMQDriver)



import uuid

from methoddispatch import SingleDispatch, singledispatch

from melange.cqrs.eventsourcing import EventStream


class AggregateRoot:
    pass


class AggregateId:
    @staticmethod
    def generate():
        return str(uuid.uuid4())


class EventSourcedAggregateRoot(AggregateRoot, SingleDispatch):
    def __init__(self, event_stream: EventStream):
        self._changes = []
        self.version = event_stream.version
        for event in event_stream.events:
            self.mutate(event)

    def add_event(self, event):
        self._changes.append(event)

    @singledispatch
    def mutate(self, event):
        pass

    def _apply(self, event):
        self.mutate(event)
        self.add_event(event)


mutator = EventSourcedAggregateRoot.mutate.register



from melange.domain_event_bus import DomainEvent


class DomainEventSerializer:
    def serialize(self, event: DomainEvent) -> str:
        raise NotImplementedError

    def deserialize(self, serialized_event: str, manifest) -> DomainEvent:
        pass


class DedupCache:
    def store(self, key, value, expire=None):
        raise NotImplementedError

    def get(self, key):
        raise NotImplementedError



from .driver_manager import DriverManager  # noqa
from .event_message import EventSchema, EventMessage  # noqa
from .event_serializer import EventSerializer  # noqa
from .exchange_listener import ExchangeListener, listener  # noqa
from .messaging_driver import Message, MessagingDriver  # noqa
from .threaded_exchange_message_consumer import ThreadedExchangeMessageConsumer  # noqa
from .exchange_message_publisher import ExchangeMessagePublisher, SQSPublisher  # noqa



from funcy import lfilter

from singleton import Singleton

from melange.cqrs.handlers import CommandHandler


class CommandBus(metaclass=Singleton):
    # TODO: For now let's make the command bus synchronous. In the future
    # we should make this async, as it is more conventional for regular
    # CQRS architectures
    def __init__(self):
        self.handlers = {}

    def register_handler(self, handler: CommandHandler):
        handled_commands = lfilter(lambda t: t is not object, handler.execute.registry)
        for handled_command in handled_commands:
            self.handlers[handled_command] = handler

    def reset(self):
        self.handlers.clear()

    def send(self, command):
        handler = self.handlers.get(type(command))
        if handler:
            return handler.execute(command)



from typing import List, Optional

from melange.domain_event_bus import DomainEvent


class EventStream:
    def __init__(self, events: List[DomainEvent], version: Optional[int]):
        self.events = events
        self.version = version

    @staticmethod
    def new(events=None):
        return EventStream(events or [], None)



import json
import uuid
from unittest.mock import MagicMock

import pytest

from melange import DriverManager
from melange.cqrs.interfaces import DomainEventSerializer
from melange.domain_event_bus import DomainEvent
from melange.messaging import MessagingDriver, ExchangeMessageConsumer, Message, \
    ExchangeListener, ExchangeMessagePublisher


class TestEventSerializer(DomainEventSerializer):
    def serialize(self, event: DomainEvent) -> str:
        pass

    def deserialize(self, serialized_event: str, manifest) -> DomainEvent:
        pass


class TestMessageConsumer:
    def test_consume_on_empty_queue(self):
        event_type_name = 'an_event_type_event'

        self._Scenario() \
            .given_an_empty_queue_to_listen() \
            .given_a_subscriber(listens_to=event_type_name) \
            .when_consuming_the_next_event_in_queue() \
            .then_the_subscriber_has_not_processed_any_event()

    def test_consume_on_queue_with_message(self):
        event_type_name = 'an_event_type_event'

        self._Scenario() \
            .given_a_queue_to_listen_with_an_event(event_of_type=event_type_name) \
            .given_a_subscriber(listens_to=event_type_name) \
            .when_consuming_the_next_event_in_queue() \
            .then_the_subscriber_has_processed_the_event() \
            .the_message_has_been_acknowledged()

    def test_consume_on_queue_with_message_but_subscriber_listens_to_different_events(self):
        event_type_name = 'an_event_type_event'
        event_type_name_2 = 'an_event_type_event_2'

        self._Scenario() \
            .given_a_queue_to_listen_with_an_event(event_of_type=event_type_name) \
            .given_a_subscriber(listens_to=event_type_name_2) \
            .when_consuming_the_next_event_in_queue() \
            .then_the_subscriber_has_not_processed_any_event() \
            .the_message_has_been_acknowledged()

    def test_unsubscribe(self):
        event_type_name = 'an_event_type_event'

        self._Scenario() \
            .given_a_queue_to_listen_with_an_event(event_of_type=event_type_name) \
            .given_a_subscriber(listens_to=event_type_name) \
            .when_unsubscribing_the_subscriber() \
            .when_consuming_the_next_event_in_queue() \
            .then_the_subscriber_has_not_processed_any_event() \
            .the_message_has_been_acknowledged()

    class _Scenario:
        def __init__(self):
            self.message_consumer = None
            self.subscriber = None
            self.event = None
            self.messages = None
            self.driver = MagicMock(spec=MessagingDriver)
            DriverManager().use_driver(driver=self.driver)

        def given_a_queue_to_listen_with_an_event(self, event_of_type=None):
            self.messages = [self._create_message(i) for i in range(1)]
            queue = self._create_queue(self.messages)
            self.driver.declare_queue.return_value = (queue, '')

            self.event = Message(uuid.uuid4(), {'event_type_name': event_of_type}, None)
            self.driver.retrieve_messages.return_value = [self.event]

            event_queue_name = 'a_queue_name'
            topic_to_subscribe = 'a_topic_name'
            self.message_consumer = ExchangeMessageConsumer(event_queue_name, topic_to_subscribe)

            return self

        def given_a_subscriber(self, listens_to=None):
            self.subscriber = MagicMock(spec=ExchangeListener)
            self.subscriber.listens_to.return_value = listens_to

            self.message_consumer.subscribe(self.subscriber)

            return self

        def when_consuming_the_next_event_in_queue(self):
            self.message_consumer.consume_event()

            return self

        def then_the_subscriber_has_processed_the_event(self):
            self.subscriber.process_event.assert_called()
            return self

        def the_message_has_been_acknowledged(self):
            self.driver.acknowledge.assert_called()

            return self

        def given_an_empty_queue_to_listen(self):
            queue = self._create_queue()
            self.driver.declare_queue = MagicMock(return_value=(queue, ''))
            self.driver.declare_topic = MagicMock()

            event_queue_name = 'a_queue_name'
            topic_to_subscribe = 'a_topic_name'
            self.message_consumer = ExchangeMessageConsumer(event_queue_name, topic_to_subscribe)

            return self

        def then_the_subscriber_has_not_processed_any_event(self):
            self.subscriber.process.assert_not_called()
            return self

        def _create_queue(self, messages=[]):
            queue = MagicMock()

            self.driver.retrieve_messages.return_value = messages

            return queue

        def _create_message(self, i):
            message = MagicMock()
            message.body = json.dumps({
                'event_type_name': 'an_event_name'
            })
            return message

        def when_unsubscribing_the_subscriber(self):
            self.message_consumer.unsubscribe(self.subscriber)
            return self


class TestMessageConsumer2:
    def setup_method(self, m):
        self.exchange_consumer = None
        DriverManager().use_driver(driver_name='aws')

    def teardown_method(self):
        driver = DriverManager().get_driver()
        if self.exchange_consumer:
            if self.exchange_consumer._event_queue:
                driver.delete_queue(self.exchange_consumer._event_queue)
            if self.exchange_consumer._dead_letter_queue:
                driver.delete_queue(self.exchange_consumer._dead_letter_queue)
            if self.exchange_consumer._topics:
                for topic in self.exchange_consumer._topics:
                    driver.delete_topic(topic)

    def test_consume_event_from_sqs(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(self._get_queue_name(),
                                                         topic_name)
        exchange_publisher = ExchangeMessagePublisher(topic=topic_name)

        self.listened_event = None

        class TestListener(ExchangeListener):
            def __init__(self):
                super().__init__(None)

            def process(x, event, **kwargs):
                self.listened_event = event

            def listens_to(x):
                return ['TestEvent']

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({'value': 'somevalue'}, event_type_name='TestEvent')

        self.exchange_consumer.consume_event()
        assert self.listened_event['value'] == 'somevalue'
        assert self.listened_event['event_type_name'] == 'TestEvent'

    def test_consume_event_with_listeners_that_listen_multiple_events(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(self._get_queue_name(),
                                                         topic_name)
        exchange_publisher = ExchangeMessagePublisher(topic=topic_name)

        self.listened_events = set()

        class TestListener(ExchangeListener):
            def __init__(self):
                super().__init__(None)

            def process(x, event, **kwargs):
                self.listened_events.add(event['value'])

            def listens_to(x):
                return ['TestEvent', 'TestEvent2', 'TestEvent3']

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({'value': 'somevalue'}, event_type_name='TestEvent')
        exchange_publisher.publish({'value': 'somevalue2', 'event_type_name': 'TestEvent2'})
        exchange_publisher.publish({'value': 'somevalue3'}, event_type_name='TestEvent4')
        exchange_publisher.publish({'value': 'somevalue4'}, event_type_name='TestEvent3')

        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()

        assert len(self.listened_events) == 3
        assert 'somevalue' in self.listened_events
        assert 'somevalue2' in self.listened_events
        assert 'somevalue4' in self.listened_events
        assert 'somevalue3' not in self.listened_events

    def _get_queue_name(self):
        return 'test_queue_{}'.format(uuid.uuid4())

    def _get_topic_name(self):
        return 'test_queue_{}'.format(uuid.uuid4())


@pytest.mark.skip()
class TestMessageConsumerRabbitMQ:
    def setup_method(self, m):
        self.exchange_consumer = None
        DriverManager().use_driver(driver_name='rabbitMQ',
                                   host='localhost')
        self.driver = DriverManager().get_driver()

    def teardown_method(self):
        if self.exchange_consumer._topics:
            for topic in self.exchange_consumer._topics:
                self.driver.delete_topic(topic)

        if self.exchange_consumer._event_queue:
            self.driver.delete_queue(self.exchange_consumer._event_queue)

        if self.exchange_consumer._dead_letter_queue:
            self.driver.delete_queue(self.exchange_consumer._dead_letter_queue)

    def test_consume_event_from_queue(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(self._get_queue_name(),
                                                         topic_name)
        exchange_publisher = ExchangeMessagePublisher(topic=topic_name)

        self.listened_event = None

        class TestListener(ExchangeListener):
            def process(x, event, **kwargs):
                self.listened_event = event

            def listens_to(self):
                return ['TestEvent']

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({'value': 'somevalue'}, event_type_name='TestEvent')

        self.exchange_consumer.consume_event()
        assert self.listened_event['value'] == 'somevalue'
        assert self.listened_event['event_type_name'] == 'TestEvent'

    def test_consume_event_with_listeners_that_listen_multiple_events(self):
        topic_name = self._get_topic_name()
        self.exchange_consumer = ExchangeMessageConsumer(self._get_queue_name(),
                                                         topic_name)
        exchange_publisher = ExchangeMessagePublisher(topic=topic_name)

        self.listened_events = set()

        class TestListener(ExchangeListener):
            def process(x, event, **kwargs):
                self.listened_events.add(event['value'])

            def listens_to(x):
                return ['TestEvent', 'TestEvent2', 'TestEvent3']

        self.exchange_consumer.subscribe(TestListener())
        exchange_publisher.publish({'value': 'somevalue'}, event_type_name='TestEvent')
        exchange_publisher.publish({'value': 'somevalue2', 'event_type_name': 'TestEvent2'})
        exchange_publisher.publish({'value': 'somevalue3'}, event_type_name='TestEvent4')
        exchange_publisher.publish({'value': 'somevalue4'}, event_type_name='TestEvent3')

        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()
        self.exchange_consumer.consume_event()

        assert len(self.listened_events) == 3
        assert 'somevalue' in self.listened_events
        assert 'somevalue2' in self.listened_events
        assert 'somevalue4' in self.listened_events
        assert 'somevalue3' not in self.listened_events

    def _get_queue_name(self):
        return 'test_queue_{}'.format(uuid.uuid4())

    def _get_topic_name(self):
        return 'test_topic_{}'.format(uuid.uuid4())



import logging

from funcy import lmap
from methoddispatch import singledispatch, SingleDispatch

from melange.cqrs.interfaces import DomainEventSerializer, DedupCache
from melange.infrastructure import Cache
from .utils import get_fully_qualified_name

logger = logging.getLogger(__name__)


class ExchangeListener(SingleDispatch):
    def __init__(self, domain_event_serializer: DomainEventSerializer, cache: DedupCache = None):
        self.cache = cache or Cache()
        self.domain_event_serializer = domain_event_serializer

    def process_event(self, event, **kwargs):
        message_listener_key = get_fully_qualified_name(self) + '.' + kwargs['message_id']
        if message_listener_key in self.cache:
            logger.info('detected a duplicated message, ignoring')
            return

        self.process(event, **kwargs)

        self.cache.store(message_listener_key, message_listener_key)

    def process(self, event, **kwargs):
        deserialized_event = self.domain_event_serializer.deserialize(event['data'], event['name'])
        self._process(deserialized_event)

    @singledispatch
    def _process(self, event):
        """Event should be an instance of DomainEvent"""
        pass

    def listens_to(self):
        accepted_events = filter(lambda t: t is not object, self._process.registry)
        return lmap(lambda ev: ev.__name__, accepted_events)

    def accepts(self, event_type_name):
        return not self.listens_to() or event_type_name in self.listens_to()


listener = ExchangeListener._process.register



import logging
import uuid

from .event_serializer import EventMessage
from .event_serializer import EventSerializer
from .driver_manager import DriverManager


logger = logging.getLogger(__name__)


class ExchangeMessagePublisher:
    def __init__(self, topic, driver=None):
        self._driver = driver or DriverManager().get_driver()
        self._topic_name = topic

    def init(self):
        self._topic = self._driver.declare_topic(self._topic_name)

    def publish(self, event, event_type_name=None):
        if not isinstance(event, EventMessage) and not isinstance(event, dict):
            logger.error('Invalid data passed. You must pass an event instance or a dict')
            raise Exception('Invalid data passed. You must pass an event instance or a dict')

        if not event_type_name and (isinstance(event, dict) and 'event_type_name' not in event):
            raise Exception('When passing an event as a dict it has to include at least the event_type_name property')

        if event_type_name and isinstance(event, dict):
            event['event_type_name'] = event_type_name

        content = EventSerializer().serialize(event)

        if not event_type_name:
            event_type_name = event.event_type_name if isinstance(event, EventMessage) else event['event_type_name']

        self.init()
        self._driver.publish(content, self._topic, event_type_name=event_type_name)

        return True


class SQSPublisher:
    def __init__(self, queue_name, dlq_name=None, driver=None, **kwargs):
        self._driver = driver or DriverManager().get_driver()
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(
            queue_name,
            dead_letter_queue_name=dlq_name)
        self.default_message_group_id = kwargs.get('message_group_id')
        self.is_fifo = self._event_queue.attributes.get('FifoQueue') == 'true'

    def publish(self, event, event_type_name=None, **kwargs):
        if isinstance(event, dict) and 'event_type_name' in event:
            event_type_name = event['event_type_name']

        if not event_type_name:
            raise Exception('You need to supply the event_type_name either in the body of the event '
                            'or as parameter')

        if isinstance(event, dict):
            event['event_type_name'] = event_type_name

        content = EventSerializer().serialize(event)

        message_group_id = kwargs.get('message_group_id', self.default_message_group_id)
        message_deduplication_id = (
            None if not self.is_fifo
            else kwargs.get('message_deduplication_id', str(uuid.uuid4()))
        )

        self._driver.queue_publish(
            content, self._event_queue, event_type_name=event_type_name,
            message_group_id=message_group_id,
            message_deduplication_id=message_deduplication_id
        )



from unittest.mock import MagicMock

from melange.messaging import MessagingDriver, DriverManager, ExchangeMessagePublisher


class TestMessagePublisher:
    def test_publish_a_message(self):
        a_topic = 'a_topic'
        event = {'some_content': '12345'}

        driver = MagicMock(spec=MessagingDriver)
        DriverManager().use_driver(driver=driver)

        message_publisher = ExchangeMessagePublisher(a_topic)
        success = message_publisher.publish(event, event_type_name='some_event')

        assert success
        driver.publish.assert_called()



import logging

from melange.messaging import EventMessage
from melange.messaging import EventSerializer
from melange.messaging import ExchangeListener
from .driver_manager import DriverManager


logger = logging.getLogger(__name__)


class ExchangeMessageYielder:
    def __init__(self, event_queue_name, *topics_to_subscribe, dead_letter_queue_name=None, driver=None, **kwargs):
        self._driver = driver or DriverManager().get_driver()
        self._event_queue_name = event_queue_name
        self._topics_to_subscribe = topics_to_subscribe
        self._dead_letter_queue_name = dead_letter_queue_name

        self._topics = [self._driver.declare_topic(t) for t in self._topics_to_subscribe]
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(
            self._event_queue_name, *self._topics,
            dead_letter_queue_name=self._dead_letter_queue_name,
            filter_events=kwargs.get('filter_events'))

    def get_messages(self):
        event_queue = self._driver.get_queue(self._event_queue_name)

        while True:
            messages = self._driver.retrieve_messages(event_queue)

            for message in messages:
                if 'event_type_name' in message.content:
                    yield EventSerializer().deserialize(message.content)

    def acknowledge(self, message):
        self._driver.acknowledge(message)


class ExchangeMessageConsumer:
    def __init__(self, event_queue_name, *topics_to_subscribe, dead_letter_queue_name=None, driver=None, **kwargs):
        self._exchange_listeners = []
        self._driver = driver or DriverManager().get_driver()
        self._event_queue_name = event_queue_name
        self._topics_to_subscribe = topics_to_subscribe
        self._dead_letter_queue_name = dead_letter_queue_name

        self._topics = [self._driver.declare_topic(t) for t in self._topics_to_subscribe]
        self._event_queue, self._dead_letter_queue = self._driver.declare_queue(
            self._event_queue_name, *self._topics,
            dead_letter_queue_name=self._dead_letter_queue_name,
            filter_events=kwargs.get('filter_events'))

    def subscribe(self, exchange_listener):
        if not isinstance(exchange_listener, ExchangeListener):
            return False

        if exchange_listener not in self._exchange_listeners:
            self._exchange_listeners.append(exchange_listener)

    def unsubscribe(self, exchange_listener):
        if exchange_listener in self._exchange_listeners:
            self._exchange_listeners.remove(exchange_listener)

    def consume_event(self):
        event_queue = self._driver.get_queue(self._event_queue_name)

        messages = self._driver.retrieve_messages(event_queue)

        for message in messages:
            try:
                self._process_message(message)
            except Exception as e:
                logger.exception(e)

    def _get_subscribers(self, event_type_name):
        return [listener for listener in self._exchange_listeners if listener.accepts(event_type_name)]

    def _process_message(self, message):
        if 'event_type_name' not in message.content:
            # The message will be ignored but not deleted from the
            # queue. Let the dead letter queue handle it
            return True

        event = EventSerializer().deserialize(message.content)
        event_type_name = event.event_type_name if isinstance(event, EventMessage) else event['event_type_name']
        subscribers = self._get_subscribers(event_type_name)

        successful = 0
        for subscr in subscribers:
            try:
                subscr.process_event(event, message_id=message.message_id)
                successful += 1
            except Exception as e:
                logger.exception(e)

        if successful == len(subscribers):
            self._driver.acknowledge(message)
